#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# coding=utf-8

#######################################
config = {
    'config':        'default.cfg',
    'teams':        ['Белые', 'Чёрные'],
    'timeout':       90,
    'gamecards':     100,
    'roundcards':    10,
    'skippenalty':   10,
    'rejpenalty':    20,
    'scoreswin':     1,
    'scoresfail':    1,
    'dicts':        ['Ushakov.txt', 'Phrases.txt'],
    'blacklist':     'blacklist.txt',
    'relax':         False,
    'showonce':      False,
    'hardtimeout':   False,
    'savegame':      True,
    'phraseword':    3,
    'phrasescores': ['4:0.5', '8:0.7'],
    }
#######################################

# ***** BEGIN GPL LICENSE BLOCK *****
#
# --------------------------------------------------------------------------
Version = 'Alias, программная реализация v0.5'
Author  = '(c) 2010 Vlad "SATtva" Miller, http://vladmiller.info'
# --------------------------------------------------------------------------
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****

import random, time, threading, math, getopt, os, pickle, base64
from collections import deque
from sys import argv, exit, getfilesystemencoding

try:
    os.chdir(os.path.dirname(os.path.abspath(__file__.encode(getfilesystemencoding()))))
except:
    pass

args   = argv[1:]
dict   = []
bl     = []
key    = None
step   = 0.2
stats  = []
help   = False
pause  = False
resume = False
maxtry = 500
shown  = 'shown.txt'
sav    = 'savegame.sav'

waves = {'gong':'gong.wav', 'ding':'ding.wav'}
keys  = {'enter':'\r', 'space':' ',
         #'backspace1':'\x08', 'backspace2':'\x7f',
         'esc':'\x1b', 'ctrl-c':'\x03', 'ctrl-f':'\x06'}

#
# Class/func definitions
#

class _GetKey():
    """Gets a single character from standard input.
        Does not echo to the screen."""
    def __init__(self):
        try:
            self.impl = _GetKeyWin()
        except ImportError:
            self.impl = _GetKeyUnix()
    def __call__(self):
        global key
        key = self.impl()
        if type(key) is bytes: key = key.decode('utf-8', 'ignore')
        return key

class _GetKeyUnix(object):
    def __init__(self):
        import tty, sys
    def __call__(self):
        import sys, tty, termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch

class _GetKeyWin(object):
    def __init__(self):
        import msvcrt
    def __call__(self):
        import msvcrt
        return msvcrt.getch()

GetKey = _GetKey()

def sound(wav):
    try:
        try:
            import winsound
            winsound.PlaySound(wav, winsound.SND_FILENAME)
        except ImportError:
            from wave import open as waveOpen
            from ossaudiodev import open as ossOpen, AFMT_U8 as format
            s = waveOpen(wav, 'rb')
            (nc, sw, fr, nf, comptype, compname) = s.getparams()
            dsp = ossOpen('/dev/dsp', 'w')
            dsp.setparameters(format, nc, fr)
            data = s.readframes(nf)
            s.close()
            dsp.write(data)
            dsp.close()
    except:
        print(" * не удалось воспроизвести '{}'".format(wav))

def pickCard():
    global dict, bl
    return random.choice(dict).strip().upper()

def scoreCard(card):
    global config
    if not config['phrasescores']:
        return config['scoreswin']

    c = []
    for n in card:
        if n.isalpha() or (n.isspace() and c and not c[-1].isspace()):
            c.append(n)

    words = 0
    for w in ''.join(c).split():
        if len(w) >= config['phraseword']:
            words += 1
    if words == 0: words = 1

    if type(config['phrasescores'][0]) is str:
        scores = {}
        for s in config['phrasescores']:
            k, v = s.split(':')
            scores[int(k)] = float(v)
        config['phrasescores'] = sorted(scores.items())

    s = config['scoreswin']
    for n in config['phrasescores']:
        if words >= n[0]:
            s = math.floor(words*n[1])
    return s

def dictFilter():
    global dict, bl
    dict = list(set(dict).difference(set(bl)))

def blWord(word, file=False):
    global config, bl
    word = word.lower()
    open(file if file else config['blacklist'], 'a', encoding='utf-8').write(word+'\n')
    bl.append(word)
    dictFilter()

def readtext(file, normalize=True):
    t = []
    f = open(file, encoding='utf-8')
    while True:
        l = f.readline()
        if normalize: l = l.lower()
        if not l: break
        elif l[0] in [' ', '#', '\t', '\n', '\r']: continue
        else: t.append(l.strip())
    return t

def parseConfig(file):
    global config
    for l in readtext(file, False):
        k, v = l.split(':')
        v = v.strip()
        if v.lower() in ['on', 'yes', 'true']: v = True
        elif v.lower() in ['off', 'no', 'false']: v = False
        elif v.isdecimal() == True: v = int(v)
        elif ',' in v:
            w = []
            for n in v.split(','): w.append(n.strip())
            v = w
        config[k.replace('-', '').strip()] = v

def saveGame():
    global sav, config, cards, teams, round, stats
    if not config['savegame']:
        return
    dump = pickle.dumps([config, cards, teams, round, stats])
    open(sav, 'wb').write(base64.b64encode(dump))

def resumeGame():
    global sav, resume, config, cards, teams, round, stats
    try:
        dump = pickle.loads(base64.b64decode(open(sav, 'rb').read()))
        resume = True
        print()
        print('Возобновление прерванной игры из файла {}...'.format(sav))
    except IOError:
        resume = False
        return
    config, cards, teams, round, stats = dump

#
# Preparing config
#

try:
    parseConfig(config['config'])
except:
    pass

if config['teams']:
    teams = [{'name':n, 'score':0} for n in config['teams']]

#
# Command line parameters
#

if args:
    opts, args = getopt.gnu_getopt(args, 'c:t:g:r:s:j:d:b:',
        ['config=', 'timeout=', 'game-cards=', 'round-cards=',
         'skip-penalty=', 'rej-penalty=', 'dicts=', 'blacklist=',
         'resume=', 'no-resume', 'help', 'debug'])
    for o,v in opts:
        long = ['--timeout', '--game-cards', '--round-cards', \
                '--skip-penalty', '--rej-penalty']
        short = ['-t', '-g', '-r', '-s', '-j']
        if o in short: o = long[short.index(o)]
        if o in long:
            config[o.replace('-', '')] = int(v)
        elif o in ['-c', '--config']:
            parseConfig(v)
            if config['teams']:
                teams = [{'name':n, 'score':0} for n in config['teams']]
        elif o in ['-d', '--dicts']:
            config['dicts'] = v.split(':')
        elif o in ['-b', '--blacklist']:
            config['blacklist'] = v
        elif o in ['--resume']:
            resume = True
            sav = v
            break
        elif o in ['--no-resume']:
            try: os.remove(sav)
            except IOError: pass
        elif o == '--help':
            help = True
        elif o == '--debug':
            print(GetKey().__repr__())
            exit(0)
    if args and not resume:
        teams = [{'name':n, 'score':0} for n in args[:10]]

resumeGame()

if not teams or help:
    print()
    print(Version)
    print(Author)
    print('По лицензии GNU GPL. Для подробностей см. README.txt')
    print()
    print('Использование: {} [<team1> [team2 [...]]] [options [...]]'.format(os.path.basename(argv[0])))
    print()
    print('Опции:')
    print('  teamN'.ljust(22), '- собственные названия играющих команд')
    print()
    print('  -t, --timeout N'.ljust(22), '- продолжительность раунда в секундах (по умолчанию {})'.format(config['timeout']))
    print('  -g, --game-cards N'.ljust(22), '- количество карт на игру (по умолчанию {})'.format(config['gamecards']))
    print('  -r, --round-cards N'.ljust(22), '- количество карт на раунд (по умолчанию {})'.format(config['roundcards']))
    print('  -s, --skip-penalty N'.ljust(22), '- штраф (сек.) за откладывание карты (по умолчанию {})'.format(config['skippenalty']))
    print('  -j, --rej-penalty N'.ljust(22), '- штраф (сек.) за смену/блэклист карты (по умолчанию {})'.format(config['rejpenalty']))
    print()
    print('  -c, --config F'.ljust(22), '- читать настройки из указанного файла (см. README.txt)')
    print('  -d, --dicts F:F:...'.ljust(22), '- путь к файлам словарей (по умолчанию задано {})'.format(len(config['dicts'])))
    print('  -b, --blacklist F'.ljust(22), '- путь к чёрному списку (по умолчанию {}),'.format(config['blacklist']))
    print('  '.ljust(22), '  пустая строка "" для отключения чёрного списка')
    print('  --resume F'.ljust(22), '- возобновить игру из указанного файла')
    print('  --no-resume'.ljust(22), '- сбросить ранее прерванную игру')
    print('  --help'.ljust(22), '- вывести данную справку')
    print('  --debug'.ljust(22), '- вывести код нажатой клавиши')
    print()
    print('Управление в игре:')
    print('    Enter'.ljust(22), '- карта отгадана, взять следующую')
    print('    Space'.ljust(22), '- отложить карту вниз, чтобы вернуться к ней позднее *')
    print('    Esc'.ljust(22), '- нарушение правил, убрать карту в колоду')
    print('    P'.ljust(22), '- поставить игру на паузу / продолжить игру')
    print('    Ctrl+F'.ljust(22), '- отправить карту в чёрный список *')
    print('    Ctrl+C'.ljust(22), '- прервать игру')
    print(' * возможно наложение штрафа')
    print()
    exit(0)

#
# Game preparations
#

print()
print('Подготовка к игре:')
print('  подключаем словари...')
for file in config['dicts']:
    try:
        dict = list(set(dict).union(set(readtext(file))))
        print('    {} - готово'.format(file))
    except IOError:
        print('    {} - ошибка при открытии словаря'.format(file))
if not dict:
    print('Не удалось подключить ни одного словаря, невозможно продолжить')
    exit(1)
else:
    print('    {} слов/фраз'.format(len(dict)))
if config['blacklist'] or config['showonce']:
    print('  подключаем чёрные списки...')
    if config['blacklist']:
        try:
            bl = list(set(bl).union(set(readtext(config['blacklist']))))
            print('    {} - готово'.format(config['blacklist']))
        except IOError:
            print('    {} - ошибка при открытии чёрного списка'.format(config['blacklist']))
    if config['blacklist']:
        try:
            bl = list(set(bl).union(set(readtext(shown))))
            print('    {} - готово'.format(shown))
        except IOError:
            pass
    print('    {} слов/фраз'.format(len(bl)))
else:
    print('  чёрные списки отключены')
if bl:
    print('  фильтрация словаря по чёрному списку...')
    dictFilter()
    print('    {} слов/фраз в словаре после фильтрации'.format(len(dict)))
# convention: left side is a top of the stack, right side is a bottom of the stack
if not resume:
    print('  отбираем {} случайных карт...'.format(config['gamecards']))
    cards = deque(maxlen=config['gamecards'])
    try:
        for i in range(config['gamecards']):
            for j in range(maxtry):
                c = pickCard()
                if config['relax'] or not c in cards:
                    cards.append(c)
                    break
    except IndexError:
        print('Все возможные слова занесены в чёрный список, невозможно продолжить')
        exit(1)
    if len(cards) < config['gamecards']:
        print('    недостаточно уникальных слов, отобрано', len(cards))
    else:
        print('    готово')
print()
print('Правила:')
print('  продолжительность раунда - {} секунд'.format(config['timeout']))
print('  {} карт в колоде,'.format(config['gamecards']), '{} карт на раунд'.format(config['roundcards']))
print('  штраф за откладывание карты - {} секунд'.format(config['skippenalty']))
print('  штраф за смену/блэклист карты - {} секунд'.format(config['rejpenalty']))
print('  выигрыш баллов за угаданное слово - {}'.format(config['scoreswin']))
print('  проигрыш баллов за нарушение правил - {}'.format(config['scoresfail']))
print()
if not resume:
    print('Участвуют команды:')
    for n in teams:
        print('  '+n['name'])
    print()
    print('Начинаем игру!')
    print()
else:
    print('Текущий счёт:')
    for team in teams:
        print('  {}:'.format(team['name']).ljust(20), team['score'])
    print()

#
# Main game routine
#

if not resume:
    round = 0
while len(cards) > 0:
    if round: saveGame()
    playing = round % len(teams)
    round += 1
    violation = False
    stats.append({'отгадано':[0, []], 'отложено':0, 'нарушений':0, 'заблэклищены':[0, []]})
    # picking cards from the main stack
    roundcards = deque(maxlen=config['roundcards'])
    for i in range(config['roundcards']):
        try:
            roundcards.append(cards.popleft())
        except IndexError:
            break
    print('Раунд {}, ход команды {} ({} карт всего, {} карт на руках)'.format(round, teams[playing]['name'], len(cards)+len(roundcards), len(roundcards)))
    print('Когда будете готовы, нажмите любую клавишу...')
    print()
    if GetKey() == keys['ctrl-c']:
        print('Игра прервана')
        print()
        exit(0)

    #
    # Round routine
    #

    t = config['timeout']
    timesup = False
    while len(roundcards) > 0:
        key  = None
        card = roundcards.popleft()
        ok = False
        while not ok:
            try:
                print('  {}  ({} / {})'.format(card, len(roundcards)+1, len(cards)+len(roundcards)+1))
                ok = True
            except:
                blWord(card, shown)  # FIXME: potential infinite loop if no more cards left in dict
                card = pickCard()
        print()
        tKey = threading.Thread(target=GetKey)
        tKey.start()

        #
        # Word routine
        #

        while True:
            print(' '*65, '\r', end='')
            if not timesup:
                m = int(t//60)
                s = math.ceil(int(t)%60)
                print('    {}:{}'.format(m, str(s).zfill(2)).rjust(7), end='')
            elif config['hardtimeout']:
                cards.append(card)
                break
            elif key is not None and not key in keys.values():
                key = GetKey()

            if key in keys.values():
                if key == keys['ctrl-c']:
                    print()
                    print('Игра прервана')
                    print()
                    exit(0)
                elif key == keys['enter']:
                    s = scoreCard(card)
                    print('  карта отгадана, +{}'.format(s))
                    print()
                    if not timesup or len(teams) < 2:
                        teams[playing]['score'] += s
                        stats[round-1]['отгадано'][0] += 1
                        stats[round-1]['отгадано'][1].append(card)
                    else:
                        print('  Укажите правильно ответившую команду:')
                        for i, team in enumerate(teams):
                            print("    {} (клавиша '{}')".format(team['name'], i+1), 'Enter' if i==playing else '')
                        ok = False
                        while not ok:
                            key = GetKey()
                            if key == keys['enter']:
                                teams[playing]['score'] += s
                                ok = True
                            elif key.isdigit() and int(key) in range(1, len(teams)+1):
                                teams[int(key)-1]['score'] += s
                                ok = True
                        stats[round-1]['отгадано'][0] += 1
                        stats[round-1]['отгадано'][1].append(card)
                    if config['showonce']:
                        blWord(card, shown)
                elif key == keys['space']:
                    print('  карта отложена', '(-{} сек.)'.format(config['skippenalty']) if config['skippenalty'] else '')
                    print()
                    roundcards.append(card)
                    t -= config['skippenalty']
                    stats[round-1]['отложено'] += 1
                elif key == keys['esc']:
                    print('  нарушение правил', ', -{}'.format(config['scoresfail']) if config['scoresfail'] else '')
                    print()
                    cards.append(card)
                    violation = True
                    teams[playing]['score'] -= config['scoresfail']
                    stats[round-1]['нарушений'] += 1
                elif key == keys['ctrl-f']:
                    if config['blacklist']:
                        print('  карта занесена в чёрный список', '(-{} сек.)'.format(config['rejpenalty']) if config['rejpenalty'] else '')
                        print()
                        blWord(card)
                        try:
                            for j in range(maxtry):
                                c = pickCard()
                                if config['relax'] or not c in cards:
                                    roundcards.appendleft(c)
                                    break
                        except IndexError:
                            print('  * не осталось карт, не занесённых в чёрный список')
                            print()
                        t -= config['rejpenalty']
                        stats[round-1]['заблэклищены'][0] += 1
                        stats[round-1]['заблэклищены'][1].append(card)
                    else:
                        print('  * чёрный список отключен')
                        print()
                        roundcards.appendleft(card)

                print('\r', ' '*65, '\r', end='')
                tKey.join()
                # lets pick up another card if there is
                # time left and no rule violations
                if not violation and not roundcards and not timesup and cards:
                    print('  взята бонусная карта')
                    roundcards.append(cards.popleft())
                break
            elif key is not None and not key in keys.values():
                key  = None
                tKey = threading.Thread(target=GetKey)
                tKey.start()

            time.sleep(step)

            if key == 'p':
                pause = True
                threading.Thread(target=sound, args=(waves['ding'],)).start()
                print()
                print()
                print('Пауза. Для продолжения нажмите P')
                print()
                while pause:
                    key = GetKey()
                    if key == 'p':
                        threading.Thread(target=sound, args=(waves['ding'],)).start()
                        pause = False
                        key   = None
                        tKey  = threading.Thread(target=GetKey)
                        tKey.start()

            if t > 0:
                t -= step
            else:
                if not timesup:
                    threading.Thread(target=sound, args=(waves['gong'],)).start()
                timesup = True

        # time's up
        if timesup:
            if config['hardtimeout']:
                print('\r', ' '*65, '\r')
                print('Время вышло, нажмите любую клавишу...')
            cards.extend(roundcards)
            print('\r', ' '*65, '\r')
            tKey.join()
            key = None
            break

    print('\n'*100)
    print('Итог раунда {}'.format(round))
    for k,v in stats[round-1].items():
        l = None
        if type(v) is list:
            l = v[1]
            v = str(v[0])
            if l:
                v += '\n    '+('\n    ').join(l)
        print('  {}:'.format(k).ljust(20), v)
    print('Командный счёт:')
    for team in teams:
        print('  {}:'.format(team['name']).ljust(20), team['score'])
    print()

#
# Game stats
#

try: os.remove(sav)
except IOError: pass

print()
print('ИТОГ ИГРЫ:')
for team in teams:
    print('  {}:'.format(team['name'].upper()).ljust(20), team['score'])
print()
GetKey()
